from ledgerblue.comm import getDongle
import struct
from solana.rpc.commitment import Finalized
from solders.pubkey import Pubkey
from solana.transaction import *
from solders.sysvar import *
from solana.rpc.async_api import AsyncClient

SOL_DERIVATION_PATH = "44'/501'/"

INS_GET_PUBKEY = 0x05
INS_SIGN_MESSAGE = 0x06

P1_NON_CONFIRM = 0x00
P1_CONFIRM = 0x01

P2_EXTEND = 0x01
P2_MORE = 0x02

MAX_CHUNK_SIZE = 255

LEDGER_CLA = 0xE0


BIP32_HARDENED_BIT = (1 << 31) >> 0


def _harden(n):
    return (n | BIP32_HARDENED_BIT) >> 0


def _extend_and_serialize_derivations_path(derivations_path: bytes):
    return (1).to_bytes(1, byteorder="little") + derivations_path


async def make_message(tx: Transaction, asyncclient: AsyncClient, find_blockhash=True):
    if find_blockhash:
        blockhash_resp = await asyncclient.get_latest_blockhash(Finalized)
        recent_blockhash = asyncclient._process_blockhash_resp(
            blockhash_resp, used_immediately=True
        )
        tx.recent_blockhash = recent_blockhash
    return tx.serialize_message()


class ledgerDongle:
    def __init__(self):
        self.dongle = getDongle(debug=False)

    def get_derive_path(self, account=None, change=None):
        length = 0
        if type(account) == int:
            if type(change) == int:
                length = 4
            else:
                length = 3
        else:
            length = 2
        derivation_path = length.to_bytes(1, byteorder="big")
        derivation_path += _harden(44).to_bytes(4, "big")
        derivation_path += _harden(501).to_bytes(4, "big")
        if length > 2:
            derivation_path += _harden(account).to_bytes(4, "big")
            if length == 4:
                derivation_path += _harden(change).to_bytes(4, "big")
        return derivation_path

    def get_address(self, account=None, change=None):
        path = self.get_derive_path(account, change)
        data = (
            struct.pack(
                ">BBBBB", LEDGER_CLA, INS_GET_PUBKEY, P1_NON_CONFIRM, 0, len(path)
            )
            + path
        )
        response = self.dongle.exchange(data)
        return Pubkey(response)

    def list(self, limit=5, page=0):
        return list(
            map(
                lambda offset: self.get_address(offset),
                range(page * limit, (page + 1) * limit),
            )
        )

    def sign(self, message, offset: int):
        assert len(message) <= 65535, "Message to sign is too long"
        derivation_path = self.get_derive_path(offset)
        header: bytes = _extend_and_serialize_derivations_path(derivation_path)

        max_size = MAX_CHUNK_SIZE - len(header)
        message_splited = [
            message[x : x + max_size] for x in range(0, len(message), max_size)
        ]
        message_splited_prefixed = [header + s for s in message_splited]

        if len(message_splited_prefixed) > 1:
            final_p2 = P2_EXTEND
            for m in message_splited_prefixed[:-1]:
                data = (
                    struct.pack(
                        ">BBBBB",
                        LEDGER_CLA,
                        INS_SIGN_MESSAGE,
                        P1_CONFIRM,
                        P2_MORE | P2_EXTEND,
                        len(m),
                    )
                    + m
                )
                self.dongle.exchange(data)
        else:
            final_p2 = 0

        data = (
            struct.pack(
                ">BBBBB",
                LEDGER_CLA,
                INS_SIGN_MESSAGE,
                P1_CONFIRM,
                final_p2,
                len(message_splited_prefixed[-1]),
            )
            + message_splited_prefixed[-1]
        )
        # print(base58.b58encode(data))
        response = self.dongle.exchange(data)
        return response



from typing import Optional
from borsh_construct import *
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from .state import get_market_program_id


InitStruct = CStruct(
    "log_level" / U8,
    "init_commission" / U8,
    "max_primary_stake" / U64,
    "nft_holders_share" / U8,
    "initial_redemption_fee" / U8,
    "is_validator_id_switchable" / Bool,
    "unit_backing" / U64,
    "redemption_fee_duration" / U32,
    "proposal_quorum" / U8,
    "creator_royalties" / U16,
    "governance_expiration_time" / U32,
    "rarities" / Vec(U16),
    "rarity_names" / Vec(String),
    "twitter_handle" / String,
    "discord_invite" / String,
    "validator_name" / String,
    "collection_uri" / String,
    "website" / String,
    "default_uri" / String,
)

InstructionEnum = Enum(
    "MintNft"
    / CStruct("switchboard_state_bump" / U8, "permission_bump" / U8, "log_level" / U8),
    "ImprintRarity" / CStruct("log_level" / U8),
    "Init" / InitStruct,
    "Redeem" / CStruct("log_level" / U8),
    "NFTWithdraw" / CStruct("cnt" / U8, "log_level" / U8),
    "ProcessRewards" / CStruct("log_level" / U8),
    "InitRebalance" / CStruct("log_level" / U8),
    "FinalizeRebalance" / CStruct("log_level" / U8),
    "UploadUris" / CStruct("uris" / Vec(String), "rarity" / U8, "log_level" / U8),
    "ResetUris" / CStruct("log_level" / U8),
    "UnDelegateNFT" / CStruct("log_level" / U8),
    "DelegateNFT" / CStruct("log_level" / U8),
    "CreateVoteAccount" / CStruct("log_level" / U8),
    "InitGovernance",
    "VoteGovernance"
    / CStruct("numeration" / U32, "vote" / Bool, "cnt" / U8, "log_level" / U8),
    "FinalizeGovernance" / CStruct("numeration" / U32, "log_level" / U8),
    "ExecuteGovernance" / CStruct("numeration" / U32, "log_level" / U8),
    "InjectTestingData" / CStruct("num_mints" / U8, "log_level" / U8),
    "FractionalizeExisting" / InitStruct,
    enum_name="InstructionEnum",
)

GovernanceType = Enum(
    "ConfigAccount",
    "ProgramUpgrade" / CStruct("buffer_account" / U8[32], "code_link" / String),
    "VoteAccountGovernance",
    enum_name="GovernanceType",
)

ConfigAccountType = Enum(
    "MaxPrimaryStake" / CStruct("value" / U64),
    "NftHolderShare" / CStruct("value" / U8),
    "InitialRedemptionFee" / CStruct("value" / U8),
    "RedemptionFeeDuration" / CStruct("value" / U32),
    "ValidatorName" / CStruct("value" / String),
    "TwitterHandle" / CStruct("value" / String),
    "DiscordInvite" / CStruct("value" / String),
    enum_name="ConfigAccountType",
)

VoteAccountGovernance = Enum(
    "ValidatorId" / CStruct("value" / U8[32]),
    "Commission" / CStruct("value" / U8),
    enum_name="VoteAccountGovernance",
)

RegistryEnum = Enum(
    "InitConfig",
    "AddProgram",
    "RemovePrograms" / CStruct("program_count" / U8),
    "AddMarketplaceProgram",
    "Reset",
    "Blank",
    enum_name="RegistryEnum",
)


SecondaryItem = CStruct(
    "cost" / U64,
    "name" / String,
    "description" / String,
)


MediationShares = CStruct(
    "buyer" / U8,
    "seller" / U8,
    "team" / U8,
)

MarketInstructionEnum = Enum(
    "List"
    / CStruct(
        "log_level" / U8,
        "authorized_withdrawer_cost" / U64,
        "mediatable_date" / U32,
        "secondary_items" / Vec(SecondaryItem),
        "description" / String,
        "validator_name" / String,
        "validator_logo_url" / String,
    ),
    "Delist"
    / CStruct(
        "log_level" / U8,
    ),
    "Buy"
    / CStruct(
        "log_level" / U8,
    ),
    "WithdrawRewards"
    / CStruct(
        "log_level" / U8,
    ),
    "RequestMediation"
    / CStruct(
        "log_level" / U8,
    ),
    "Mediate"
    / CStruct(
        "log_level" / U8,
        "mediation_shares" / MediationShares,
    ),
    "ValidateSecondaryItemsTransfers"
    / CStruct(
        "log_level" / U8,
        "item_index" / U32,
    ),
    enum_name="InstructionEnum",
)


def build_market_instruction(instruction, accounts) -> Instruction:
    return Instruction(
        program_id=get_market_program_id(),
        data=MarketInstructionEnum.build(instruction),
        accounts=accounts,
    )


Purchase = CStruct(
    "buyer" / U8[32],
    "date" / U32,
    "date_finalized" / Option(U32),
)

StoredSecondaryItem = CStruct(
    "cost" / U64,
    "name" / String,
    "description" / String,
    "date_validated" / Option(U32),
)

Storage = CStruct(
    "validation_phrase" / U32,
    "authorized_withdrawer" / U8[32],
    "vote_account" / U8[32],
    "authorized_withdrawer_cost" / U64,
    "mediatable_date" / U32,
    "purchase" / Option(Purchase),
    "request_mediation_date" / Option(U32),
    "mediation_date" / Option(U32),
    "mediation_shares" / Option(MediationShares),
    "secondary_items" / Vec(StoredSecondaryItem),
    "description" / String,
    "validator_name" / String,
    "validator_logo_url" / String,
)


def build_governance_type(
    governance_type: GovernanceType.enum,
    config_account_type: Optional[ConfigAccountType.enum] = None,
    vote_account_governance: Optional[VoteAccountGovernance.enum] = None,
):
    if governance_type == GovernanceType.enum.ConfigAccount():
        return GovernanceType.build(governance_type) + ConfigAccountType.build(
            config_account_type
        )
    elif governance_type == GovernanceType.enum.VoteAccountGovernance():
        return GovernanceType.build(governance_type) + VoteAccountGovernance.build(
            vote_account_governance
        )
    else:
        return GovernanceType.build(governance_type)


def build_instruction(
    instruction: InstructionEnum.enum,
    title: Optional[str] = None,
    description: Optional[str] = None,
    governance_type: Optional[GovernanceType.enum] = None,
    config_account_type: Optional[ConfigAccountType.enum] = None,
    vote_account_governance: Optional[VoteAccountGovernance.enum] = None,
    log_level: int = 0,
):
    if instruction == InstructionEnum.enum.InitGovernance():
        return (
            InstructionEnum.build(instruction)
            + build_governance_type(
                governance_type,
                config_account_type=config_account_type,
                vote_account_governance=vote_account_governance,
            )
            + String.build(title)
            + String.build(description)
            + (log_level).to_bytes(1, "big")
        )
    else:
        return InstructionEnum.build(instruction)


class ComputeBudgetInstruction:
    def __init__(self):
        self.InstructionEnum = Enum(
            "RequestUnitsDeprecated" / CStruct("units" / U32, "additional_fee" / U32),
            "RequestHeapFrame" / CStruct("value" / U32),
            "SetComputeUnitLimit" / CStruct("value" / U32),
            "SetComputeUnitPrice" / CStruct("value" / U64),
            enum_name="InstructionEnum",
        )
        self.program_id = Pubkey.from_string(
            "ComputeBudget111111111111111111111111111111"
        )

    def request_heap_frame(self, total_bytes, payer) -> Instruction:
        instruction_bytes = self.InstructionEnum.build(
            self.InstructionEnum.enum.RequestHeapFrame(total_bytes)
        )
        return Instruction(
            accounts=[AccountMeta(payer, True, False)],
            program_id=self.program_id,
            data=instruction_bytes,
        )

    def set_compute_unit_limit(self, units, payer) -> Instruction:
        instruction_bytes = self.InstructionEnum.build(
            self.InstructionEnum.enum.SetComputeUnitLimit(units)
        )
        return Instruction(
            accounts=[AccountMeta(payer, True, False)],
            program_id=self.program_id,
            data=instruction_bytes,
        )

    def set_compute_unit_price(self, micro_lamports, payer) -> Instruction:
        instruction_bytes = self.InstructionEnum.build(
            self.InstructionEnum.enum.SetComputeUnitPrice(micro_lamports)
        )
        return Instruction(
            accounts=[AccountMeta(payer, True, False)],
            program_id=self.program_id,
            data=instruction_bytes,
        )



import json
from typing import Optional
import base58
from borsh_construct import *
from .ledger import *
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solana.rpc import types
from solana.rpc.async_api import AsyncClient
from solana.transaction import Transaction
from solders.rpc.responses import SendTransactionResp
from solders.signature import Signature
import os
from construct import Adapter


class Constants:
    INGL_CONFIG_SEED = "ingl_config"
    URIS_ACCOUNT_SEED = "uris_account"
    GENERAL_ACCOUNT_SEED = "general_account"
    INGL_NFT_COLLECTION_KEY = "ingl_nft_collection"
    INGL_MINT_AUTHORITY_KEY = "ingl_mint_authority"
    COLLECTION_HOLDER_KEY = "collection_holder"
    VOTE_ACCOUNT_KEY = "vote_account"
    AUTHORIZED_WITHDRAWER_KEY = "authorized_withdrawer"
    STAKE_ACCOUNT_KEY = "stake_account"
    PD_POOL_ACCOUNT_KEY = "pd_pool_account"
    NFT_ACCOUNT_CONST = "nft_account"
    INGL_PROGRAM_AUTHORITY_KEY = "ingl_program_authority"
    INGL_PROPOSAL_KEY = "ingl_proposal"
    VALIDATOR_ID_SEED = "validator_ID___________________"
    T_STAKE_ACCOUNT_KEY = "t_stake_account_key"
    T_WITHDRAW_KEY = "t_withdraw_key"
    INGL_REGISTRY_CONFIG_SEED = "config"

    PDA_AUTHORIZED_WITHDRAWER_SEED = b"authorized_withdrawer"
    PROGRAM_STORAGE_SEED = b"program_storage"
    PDA_UPGRADE_AUTHORITY_SEED = b"upgrade_authority"
    ESCROW_ACCOUNT_SEED = b"escrow_account"
    REGISTRY_STORAGE_SEED = b"marketplace_storage"

    ESCROWED_BASIS_POINTS = 2000
    TEAM_FEES_BASIS_POINTS = 10

    TEAM_ADDRESS = Pubkey.from_string("Et2tm6NsfBZJbEYXtWTv9k51V4tWtQvufexSgXoDRGVA")
    MEDIATORS = [Pubkey.from_string("Et2tm6NsfBZJbEYXtWTv9k51V4tWtQvufexSgXoDRGVA")]

    STAKE_PROGRAM_ID = Pubkey.from_string("Stake11111111111111111111111111111111111111")
    STAKE_CONFIG_PROGRAM_ID = Pubkey.from_string(
        "StakeConfig11111111111111111111111111111111"
    )
    VOTE_PROGRAM_ID = Pubkey.from_string("Vote111111111111111111111111111111111111111")
    BPF_LOADER_UPGRADEABLE = Pubkey.from_string(
        "BPFLoaderUpgradeab1e11111111111111111111111"
    )
    REGISTRY_PROGRAM = Pubkey.from_string(
        "38pfsot7kCZkrttx1THEDXEz4JJXmCCcaDoDieRtVuy5"
    )
    COMPUTE_BUDGET_PROGRAM = Pubkey.from_string(
        "ComputeBudget111111111111111111111111111111"
    )


VoteReward = CStruct(
    "epoch_number" / U64,
    "total_reward" / U64,
    "total_stake" / U32,
    "nft_holders_reward" / U64,
)

UpgradeVote = CStruct(
    "vote" / Bool,
    "validator_id" / U8[32],
)

ValidatorConfig = CStruct(
    "validation_phrase" / U32,
    "is_validator_id_switchable" / Bool,
    "max_primary_stake" / U64,
    "nft_holders_share" / U8,
    "initial_redemption_fee" / U8,
    "unit_stake" / U64,
    "redemption_fee_duration" / U32,
    "proposal_quorum" / U8,
    "creator_royalties" / U16,
    "commission" / U8,
    "validator_id" / U8[32],
    "vote_account" / U8[32],
    "governance_expiration_time" / U32,
    "default_uri" / String,
    "validator_name" / String,
    "twitter_handle" / String,
    "discord_invite" / String,
    "website" / String,
)
RebalancingData = CStruct(
    "pending_validator_rewards" / U64,
    "unclaimed_validator_rewards" / U64,
    "is_rebalancing_active" / Bool,
)

GeneralData = CStruct(
    "validation_phrase" / U32,
    "mint_numeration" / U32,
    "pending_delegation_total" / U64,
    "dealloced" / U64,
    "total_delegated" / U64,
    "last_withdraw_epoch" / U64,
    "last_total_staked" / U64,
    "is_t_stake_initialized" / Bool,
    "proposal_numeration" / U32,
    "last_feeless_redemption_date" / U32,
    "last_validated_validator_id_proposal" / U32,
    "rebalancing_data" / RebalancingData,
    "unfinalized_proposals" / HashSet(U32),
    "vote_rewards" / Vec(VoteReward),
)

NftData = CStruct(
    "validation_phrase" / U32,
    "rarity" / Option(U8),
    "rarity_seed_time" / Option(U32),
    "funds_location" / U8,
    "numeration" / U32,
    "date_created" / U32,
    "last_withdrawal_epoch" / Option(U64),
    "last_delegation_epoch" / Option(U64),
    "all_withdraws" / Vec(U64),
    "all_votes" / HashMap(U32, Bool),
)

RegistryConfig = CStruct(
    "validation_phase" / U32,
    "validator_numeration" / U32,
)


class rpc_url:
    DEVNET = "https://api.devnet.solana.com"
    TESTNET = "https://api.testnet.solana.com"
    MAINNET = "https://api.mainnet.solana.com"

    target_network = DEVNET


def get_explorer_suffix(cluster_url: str):
    if "net.solana.com" not in cluster_url:
        return "?cluster=custom&customUrl=" + cluster_url
    if cluster_url == rpc_url.DEVNET:
        return "?cluster=devnet"
    elif cluster_url == rpc_url.TESTNET:
        return "?cluster=testnet"
    else:
        return ""


def get_network_url(network: str) -> str:
    if network == "devnet":
        return rpc_url.DEVNET
    elif network == "testnet":
        return rpc_url.TESTNET
    elif network == "mainnet":
        return rpc_url.MAINNET
    else:
        raise Exception("Invalid network")


def keypair_from_json(filepath) -> Keypair:
    keypair = Keypair.from_bytes(json.load(open(filepath)))
    return keypair


class KeypairInput:
    def __init__(
        self,
        t_keypair: Optional[Keypair] = None,
        ledger_address: Optional[int] = None,
        pubkey: Optional[Pubkey] = None,
    ):
        assert (
            t_keypair or ledger_address
        ), "KeypairInput must have at least one of keypair or ledger_address"
        self.keypair = t_keypair
        self.ledger_address = ledger_address
        self.pubkey = (
            pubkey
            if pubkey
            else t_keypair.pubkey()
            if t_keypair
            else Pubkey.new_unique()
        )

    def __str__(self):
        return f"KeypairInput(keypair={self.keypair}, ledger_address={self.ledger_address}, pubkey={self.pubkey})"

    def __repr__(self):
        return self.__str__()


def parse_keypair_input(str_input: str) -> KeypairInput:
    if str_input.startswith("Ledger://"):
        t_dongle = ledgerDongle()
        pub_key = t_dongle.get_address(int(str_input[9:]))
        return KeypairInput(ledger_address=int(str_input[9:]), pubkey=pub_key)
    else:
        t_keypair = keypair_from_json(str_input)
        return KeypairInput(t_keypair=t_keypair)


class PubkeyInput:
    def __init__(
        self,
        t_keypair: Optional[Keypair] = None,
        pubkey: Optional[Pubkey] = None,
        ledger_address: Optional[int] = None,
    ):
        assert (
            t_keypair or pubkey or ledger_address
        ), "PubkeyInput must have at least one of keypair, pubkey or ledger_address"
        self.keypair = t_keypair
        self.pubkey = t_keypair.pubkey() if t_keypair else pubkey
        self.ledger_address = ledger_address

    def __str__(self):
        return f"PubkeyInput(keypair={self.keypair}, pubkey={self.pubkey}, ledger_address={self.ledger_address})"

    def __repr__(self):
        return self.__str__()


def parse_pubkey_input(str_input: str) -> PubkeyInput:
    if str_input.startswith("Ledger://"):
        t_dongle = ledgerDongle()
        pub_key = t_dongle.get_address(int(str_input[9:]))
        return PubkeyInput(ledger_address=int(str_input[9:]), pubkey=pub_key)
    else:
        try:
            pubkey = PubkeyInput(pubkey=Pubkey.from_string(str_input))
            return pubkey
        except Exception as e:
            try:
                t_keypair = keypair_from_json(str_input)
                return PubkeyInput(t_keypair=t_keypair)
            except Exception as new_e:
                print("invalid public key input")
                raise new_e


async def sign_and_send_tx(
    tx: Transaction, client: AsyncClient, *args, skip_preflight: bool = False
) -> SendTransactionResp:
    last_valid_block_height = None
    if client.blockhash_cache:
        try:
            recent_blockhash = client.blockhash_cache.get()
        except ValueError:
            blockhash_resp = await client.get_latest_blockhash(Finalized)
            recent_blockhash = client._process_blockhash_resp(
                blockhash_resp, used_immediately=True
            )
            last_valid_block_height = blockhash_resp.value.last_valid_block_height
    else:
        blockhash_resp = await client.get_latest_blockhash(Finalized)
        recent_blockhash = client.parse_recent_blockhash(blockhash_resp)
        last_valid_block_height = blockhash_resp.value.last_valid_block_height
    tx.recent_blockhash = recent_blockhash

    # print("signing actually args: ", args)
    for arg in args:
        # print(arg)
        if isinstance(arg, KeypairInput):
            # print("p_key: ", arg.pubkey, "keypair: ", arg.keypair, "ledger: ", arg.ledger_address)
            if arg.keypair is not None:
                tx.sign_partial(arg.keypair)
            elif arg.ledger_address is not None:
                t_dongle = ledgerDongle()
                message = await make_message(tx, client, False)
                # print("message: ", message)
                signature = Signature.from_bytes(
                    t_dongle.sign(message, arg.ledger_address)
                )
                tx.add_signature(arg.pubkey, signature)
            else:
                raise Exception("KeypairInput is not valid")
        else:
            raise ValueError(
                "Invalid argument expected a KeypairInput, Found -> : " + str(type(arg))
            )
    # print("Reached here")
    opts_to_use = types.TxOpts(
        preflight_commitment=client._commitment,
        last_valid_block_height=last_valid_block_height,
        skip_preflight=skip_preflight,
    )
    txn_resp = await client.send_raw_transaction(tx.serialize(), opts=opts_to_use)
    if client.blockhash_cache:
        blockhash_resp = await client.get_latest_blockhash(Finalized)
        client._process_blockhash_resp(blockhash_resp, used_immediately=False)
    # print("finished")
    return txn_resp


async def sign_send_confirm_return_tx_as_link(
    tx: Transaction, client: AsyncClient, *args
) -> str:
    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    try:
        t_dets = await sign_and_send_tx(tx, client, *args)
        await client.confirm_transaction(
            tx_sig=t_dets.value, commitment="finalized", sleep_seconds=0.4
        )

        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


def set_config(key: str, value: str):
    file_dir = f"{os.path.expanduser('~')}/.config/solana/ingl/"
    os.makedirs(file_dir, exist_ok=True)
    file_dir = file_dir + "config.json"
    try:
        f = open(file_dir, "r")
        config = json.load(f)
        f.close()
    except:
        config = {}
    config[key] = value
    with open(file_dir, "w") as f:
        json.dump(config, f)


def get_config(key: str) -> str:
    file_dir = f"{os.path.expanduser('~')}/.config/solana/ingl/"
    os.makedirs(file_dir, exist_ok=True)
    file_dir = file_dir + "config.json"
    try:
        f = open(file_dir, "r")
        config = json.load(f)
        f.close()
        if key in config:
            return config[key]
        else:
            return ""
    except:
        return ""


def get_program_id() -> Pubkey:
    program_id_str = get_config("program_id")
    try:
        return Pubkey.from_string(program_id_str)
    except:
        return Pubkey.from_string("HD8kYhgqmZCJ881vyBQ3fR6a62YL7cZBnYj1P7oLw8An")


def set_program_id(program_id: str):
    set_config("program_id", program_id)


def get_network() -> str:
    network = get_config("network")
    if network == "":
        return get_network_url(network="devnet")
    else:
        return network


def set_network(network: str):
    set_config("network", network)


def get_keypair_path() -> str:
    keypair_path = get_config("keypair_path")
    if keypair_path == "":
        return f"{os.path.expanduser('~')}/.config/solana/id.json"
    else:
        return keypair_path


def set_keypair_path(keypair_path: str) -> bool:
    path = os.path.abspath(keypair_path)
    try:
        keypair = keypair_from_json(path)
    except:
        print("Invalid keypair path")
        return False
    set_config("keypair_path", path)
    print("Keypair path set to: ", path)
    print("Keypair Public Key: ", keypair.pubkey())
    return True


def get_market_program_id() -> Pubkey:
    program_id_str = get_config("markets_program_id")
    try:
        return Pubkey.from_string(program_id_str)
    except:
        raise Exception(
            "Invalid markets program id, please set one with: ingl markets config set -p <program_id>"
        )


def set_market_program_id(program_id: str):
    set_config("markets_program_id", program_id)


def get_market_keypair_path() -> str:
    keypair_path = get_config("markets_keypair_path")
    if keypair_path == "":
        return f"{os.path.expanduser('~')}/.config/solana/id.json"
    else:
        return keypair_path


def set_market_keypair_path(keypair_path: str) -> bool:
    path = os.path.abspath(keypair_path)
    try:
        keypair = keypair_from_json(path)
    except:
        print("Invalid keypair path")
        return False
    set_config("markets_keypair_path", path)
    print("Keypair path set to: ", path)
    print("Keypair Public Key: ", keypair.pubkey())
    return True



import solders
from solders.pubkey import Pubkey
from solders import system_program
from solana.transaction import *
from spl.token import constants as spl_constants
from spl.token import instructions as assoc_instructions
from .instruction import *
from .state import *
from .state import Constants as ingl_constants
from solana.rpc.async_api import AsyncClient
from solana.rpc.api import Client
from rich import print


async def ingl_init(
    payer_keypair: KeypairInput,
    upgrade_authority: KeypairInput,
    validator_pubkey: PubkeyInput,
    init_commission: int,
    max_primary_stake: int,
    nft_holders_share: int,
    initial_redemption_fee: int,
    is_validator_id_switchable: bool,
    unit_backing: int,
    redemption_fee_duration: int,
    proposal_quorum: int,
    creator_royalties: int,
    governance_expiration_time: int,
    rarities: List[int],
    rarity_names: List[str],
    twitter_handle: str,
    discord_invite: str,
    validator_name: str,
    collection_uri: str,
    website: str,
    default_uri: str,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    mint_pubkey, _mint_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_NFT_COLLECTION_KEY, "UTF-8")], get_program_id()
    )
    mint_authority_pubkey, _mint_authority_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_MINT_AUTHORITY_KEY, "UTF-8")], get_program_id()
    )
    (
        collection_holder_pubkey,
        _collection_holder_pubkey_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.COLLECTION_HOLDER_KEY, "UTF-8")], get_program_id()
    )
    mint_associated_account_pubkey = assoc_instructions.get_associated_token_address(
        collection_holder_pubkey, mint_pubkey
    )
    metaplex_program_id = Pubkey.from_string(
        "metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s"
    )
    metadata_pda, _metadata_pda_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(mint_pubkey)],
        metaplex_program_id,
    )
    master_edition_pda, _master_edition_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(mint_pubkey), b"edition"],
        metaplex_program_id,
    )
    ingl_config_pubkey, _ingl_config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    uris_account_pubkey, _uris_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.URIS_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    this_program_data_pubkey, _this_program_data_bump = Pubkey.find_program_address(
        [bytes(get_program_id())], ingl_constants.BPF_LOADER_UPGRADEABLE
    )

    # registry_program_config_key, _registry_program_config_bump = Pubkey.find_program_address([b'config'], ingl_constants.REGISTRY_PROGRAM_ID)
    storage_key, _storage_bump = Pubkey.find_program_address(
        [b"storage"], ingl_constants.REGISTRY_PROGRAM
    )
    name_storage_meta = AccountMeta(
        Pubkey.find_program_address(
            [
                b"name_storage",
            ],
            ingl_constants.REGISTRY_PROGRAM,
        )[0],
        False,
        True,
    )
    upgrade_authority_meta = AccountMeta(upgrade_authority.pubkey, True, False)

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    collection_holder_meta = AccountMeta(collection_holder_pubkey, False, True)
    mint_account_meta = AccountMeta(mint_pubkey, False, True)
    mint_authority_meta = AccountMeta(mint_authority_pubkey, False, False)
    mint_associated_meta = AccountMeta(mint_associated_account_pubkey, False, True)
    spl_program_meta = AccountMeta(spl_constants.TOKEN_PROGRAM_ID, False, False)
    sysvar_rent_account_meta = AccountMeta(solders.sysvar.RENT, False, False)
    system_program_meta = AccountMeta(system_program.ID, False, False)
    this_program_data_meta = AccountMeta(this_program_data_pubkey, False, False)
    token_metadata_meta = AccountMeta(metadata_pda, False, True)
    metadata_program_id = AccountMeta(metaplex_program_id, False, False)
    associated_program_meta = AccountMeta(
        spl_constants.ASSOCIATED_TOKEN_PROGRAM_ID, False, False
    )
    edition_meta = AccountMeta(master_edition_pda, False, True)
    ingl_config_meta = AccountMeta(ingl_config_pubkey, False, True)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    uris_account_meta = AccountMeta(uris_account_pubkey, False, True)
    validator_account_meta = AccountMeta(validator_pubkey.pubkey, False, True)
    # registry_program_config_meta = AccountMeta(registry_program_config_key, False, True)
    program_meta = AccountMeta(get_program_id(), False, False)
    team_account_meta = AccountMeta(ingl_constants.TEAM_ADDRESS, False, True)
    storage_account_meta = AccountMeta(storage_key, False, True)
    registry_program_meta = AccountMeta(ingl_constants.REGISTRY_PROGRAM, False, False)

    accounts = [
        payer_account_meta,
        ingl_config_meta,
        general_account_meta,
        uris_account_meta,
        sysvar_rent_account_meta,
        validator_account_meta,
        collection_holder_meta,
        mint_account_meta,
        mint_authority_meta,
        mint_associated_meta,
        token_metadata_meta,
        edition_meta,
        spl_program_meta,
        system_program_meta,
        this_program_data_meta,
        upgrade_authority_meta,
        program_meta,
        team_account_meta,
        storage_account_meta,
        name_storage_meta,
        # registry_program_config_meta,
        system_program_meta,
        # programs below
        associated_program_meta,
        spl_program_meta,
        metadata_program_id,
        registry_program_meta,
    ]
    # print(accounts)
    data = build_instruction(
        InstructionEnum.enum.Init(
            init_commission=init_commission,
            max_primary_stake=max_primary_stake,
            nft_holders_share=nft_holders_share,
            initial_redemption_fee=initial_redemption_fee,
            is_validator_id_switchable=is_validator_id_switchable,
            unit_backing=unit_backing,
            redemption_fee_duration=redemption_fee_duration,
            proposal_quorum=proposal_quorum,
            creator_royalties=creator_royalties,
            governance_expiration_time=governance_expiration_time,
            rarities=rarities,
            rarity_names=rarity_names,
            twitter_handle=twitter_handle,
            discord_invite=discord_invite,
            validator_name=validator_name,
            collection_uri=collection_uri,
            website=website,
            default_uri=default_uri,
            log_level=log_level,
        )
    )
    transaction = Transaction()
    # print(data)
    transaction.add(
        ComputeBudgetInstruction().set_compute_unit_limit(300_000, payer_keypair.pubkey)
    )
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )

    try:
        t_dets = await sign_and_send_tx(
            transaction, client, payer_keypair, skip_preflight=False
        )
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def mint_nft(
    payer_keypair: KeypairInput,
    mint_keypair: KeypairInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    mint_authority_pubkey, _mint_authority_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_MINT_AUTHORITY_KEY, "UTF-8")], get_program_id()
    )
    collection_mint_pubkey, _collection_mint_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_NFT_COLLECTION_KEY, "UTF-8")], get_program_id()
    )
    pd_pool_pubkey, _pd_pool_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.PD_POOL_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    mint_associated_account_pubkey = assoc_instructions.get_associated_token_address(
        payer_keypair.pubkey, mint_keypair.pubkey
    )
    metaplex_program_id = Pubkey.from_string(
        "metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s"
    )
    metadata_pda, _metadata_pda_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(mint_keypair.pubkey)],
        metaplex_program_id,
    )
    collection_master_edition_pda, _master_edition_bump = Pubkey.find_program_address(
        [
            b"metadata",
            bytes(metaplex_program_id),
            bytes(collection_mint_pubkey),
            b"edition",
        ],
        metaplex_program_id,
    )
    mint_edition_pda, _mint_edition_bump = Pubkey.find_program_address(
        [
            b"metadata",
            bytes(metaplex_program_id),
            bytes(mint_keypair.pubkey),
            b"edition",
        ],
        metaplex_program_id,
    )
    collection_account_pda, _collection_account_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(collection_mint_pubkey)],
        metaplex_program_id,
    )
    nft_account_pubkey, _nft_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint_keypair.pubkey)],
        get_program_id(),
    )
    ingl_config_pubkey, _ingl_config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    uri_account_pubkey, _uri_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.URIS_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    mint_account_meta = AccountMeta(mint_keypair.pubkey, True, True)
    pd_pool_meta = AccountMeta(pd_pool_pubkey, False, True)
    mint_authority_meta = AccountMeta(mint_authority_pubkey, False, True)
    mint_associated_meta = AccountMeta(mint_associated_account_pubkey, False, True)
    spl_program_meta = AccountMeta(spl_constants.TOKEN_PROGRAM_ID, False, False)
    sysvar_rent_account_meta = AccountMeta(solders.sysvar.RENT, False, False)
    system_program_meta = AccountMeta(system_program.ID, False, False)
    token_metadata_meta = AccountMeta(metadata_pda, False, True)
    metadata_program_id = AccountMeta(metaplex_program_id, False, False)
    associated_program_meta = AccountMeta(
        spl_constants.ASSOCIATED_TOKEN_PROGRAM_ID, False, False
    )
    nft_account_meta = AccountMeta(nft_account_pubkey, False, True)
    collection_master_edition_meta = AccountMeta(
        collection_master_edition_pda, False, True
    )
    mint_edition_meta = AccountMeta(mint_edition_pda, False, True)
    collection_mint_meta = AccountMeta(collection_mint_pubkey, False, True)
    collection_account_meta = AccountMeta(collection_account_pda, False, True)
    ingl_config_meta = AccountMeta(ingl_config_pubkey, False, False)
    uri_account_meta = AccountMeta(uri_account_pubkey, False, True)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)

    accounts = [
        payer_account_meta,
        mint_account_meta,
        mint_authority_meta,
        mint_associated_meta,
        spl_program_meta,
        sysvar_rent_account_meta,
        system_program_meta,
        token_metadata_meta,
        pd_pool_meta,
        nft_account_meta,
        collection_master_edition_meta,
        mint_edition_meta,
        collection_mint_meta,
        collection_account_meta,
        ingl_config_meta,
        uri_account_meta,
        general_account_meta,
        system_program_meta,
        spl_program_meta,
        system_program_meta,
        spl_program_meta,
        associated_program_meta,
        spl_program_meta,
        metadata_program_id,
        metadata_program_id,
        spl_program_meta,
        metadata_program_id,
    ]

    instruction_data = build_instruction(
        InstructionEnum.enum.MintNft(
            switchboard_state_bump=0, permission_bump=0, log_level=log_level
        )
    )
    transaction = Transaction()
    transaction.add(
        ComputeBudgetInstruction().set_compute_unit_limit(400_000, payer_keypair.pubkey)
    )
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(
            transaction, client, payer_keypair, mint_keypair
        )
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def delegate_nft(
    payer_keypair: KeypairInput,
    mint_pubkey: PubkeyInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    nft_account_pubkey, _nft_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint_pubkey.pubkey)],
        get_program_id(),
    )
    mint_associated_account_pubkey = assoc_instructions.get_associated_token_address(
        payer_keypair.pubkey, mint_pubkey.pubkey
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    mint_account_meta = AccountMeta(mint_pubkey.pubkey, False, False)
    nft_account_meta = AccountMeta(nft_account_pubkey, False, True)
    mint_associated_meta = AccountMeta(mint_associated_account_pubkey, False, False)
    config_account_meta = AccountMeta(
        pubkey=config_account_pubkey, is_signer=False, is_writable=False
    )
    general_account_meta = AccountMeta(
        pubkey=general_account_pubkey, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        mint_account_meta,
        nft_account_meta,
        mint_associated_meta,
        general_account_meta,
    ]

    # print(accounts)
    instruction_data = build_instruction(
        InstructionEnum.enum.DelegateNFT(log_level=log_level)
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def undelegate_nft(
    payer_keypair: KeypairInput,
    mint_pubkey: PubkeyInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    nft_account_pubkey, _nft_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint_pubkey.pubkey)],
        get_program_id(),
    )
    mint_associated_account_pubkey = assoc_instructions.get_associated_token_address(
        payer_keypair.pubkey, mint_pubkey.pubkey
    )
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    config_data = ValidatorConfig.parse(
        (await client.get_account_info(config_account_pubkey)).value.data
    )
    expected_vote_pubkey = Pubkey(config_data.vote_account)

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    mint_account_meta = AccountMeta(mint_pubkey.pubkey, False, False)
    nft_account_meta = AccountMeta(nft_account_pubkey, False, True)
    mint_associated_meta = AccountMeta(mint_associated_account_pubkey, False, False)
    config_account_meta = AccountMeta(
        pubkey=config_account_pubkey, is_signer=False, is_writable=False
    )
    general_account_meta = AccountMeta(
        pubkey=general_account_pubkey, is_signer=False, is_writable=True
    )

    vote_account_meta = AccountMeta(expected_vote_pubkey, False, False)
    authorized_withdrawer_meta = AccountMeta(authorized_withdrawer_key, False, False)
    system_program_meta = AccountMeta(system_program.ID, False, False)

    accounts = [
        payer_account_meta,
        vote_account_meta,
        config_account_meta,
        mint_account_meta,
        nft_account_meta,
        mint_associated_meta,
        general_account_meta,
        system_program_meta,
        authorized_withdrawer_meta,
    ] + interested_proposals_account_metas

    # print(accounts)
    instruction_data = build_instruction(
        InstructionEnum.enum.UnDelegateNFT(
            log_level=log_level,
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def create_vote_account(
    validator_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    expected_vote_pubkey, _expected_vote_pubkey_nonce = Pubkey.find_program_address(
        [bytes(ingl_constants.VOTE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    expected_stake_key, _expected_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    pd_pool_account_pubkey, _pd_pool_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.PD_POOL_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )

    print(f"Vote_Account: {expected_vote_pubkey}")

    rent_account_meta = AccountMeta(solders.sysvar.RENT, False, False)
    sysvar_clock_meta = AccountMeta(solders.sysvar.CLOCK, False, False)
    validator_meta = AccountMeta(validator_keypair.pubkey, True, True)
    vote_account_meta = AccountMeta(expected_vote_pubkey, False, True)
    sys_program_meta = AccountMeta(system_program.ID, False, False)
    vote_program_meta = AccountMeta(ingl_constants.VOTE_PROGRAM_ID, False, False)
    spl_program_meta = AccountMeta(spl_constants.TOKEN_PROGRAM_ID, False, False)
    stake_account_meta = AccountMeta(expected_stake_key, False, True)
    stake_program_meta = AccountMeta(ingl_constants.STAKE_PROGRAM_ID, False, False)
    config_account_meta = AccountMeta(config_account_pubkey, False, True)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    pd_pool_account_meta = AccountMeta(pd_pool_account_pubkey, False, False)
    stake_history_meta = AccountMeta(STAKE_HISTORY, False, False)
    stake_config_meta = AccountMeta(
        ingl_constants.STAKE_CONFIG_PROGRAM_ID, False, False
    )

    accounts = [
        validator_meta,
        vote_account_meta,
        rent_account_meta,
        sysvar_clock_meta,
        spl_program_meta,
        stake_account_meta,
        config_account_meta,
        general_account_meta,
        pd_pool_account_meta,
        stake_history_meta,
        stake_config_meta,
        sys_program_meta,
        vote_program_meta,
        vote_program_meta,
        sys_program_meta,
        stake_program_meta,
    ]

    data = InstructionEnum.build(
        InstructionEnum.enum.CreateVoteAccount(
            log_level=log_level,
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, validator_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def init_rebalance(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    expected_stake_key, _expected_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_stake_key, _t_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_withdraw_key, _t_withdraw_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_WITHDRAW_KEY, "UTF-8")], get_program_id()
    )
    pd_pool_pubkey, _pd_pool_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.PD_POOL_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )


async def init_rebalance(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    expected_stake_key, _expected_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_stake_key, _t_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_withdraw_key, _t_withdraw_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_WITHDRAW_KEY, "UTF-8")], get_program_id()
    )
    pd_pool_pubkey, _pd_pool_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.PD_POOL_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )

    config_data = ValidatorConfig.parse(
        (await client.get_account_info(config_account_pubkey)).value.data
    )
    expected_vote_pubkey = Pubkey(config_data.vote_account)

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    rent_account_meta = AccountMeta(solders.sysvar.RENT, False, False)
    sysvar_clock_meta = AccountMeta(solders.sysvar.CLOCK, False, False)
    sys_program_meta = AccountMeta(system_program.ID, False, False)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    stake_account_meta = AccountMeta(expected_stake_key, False, True)
    t_stake_meta = AccountMeta(t_stake_key, False, True)
    t_withdraw_meta = AccountMeta(t_withdraw_key, False, True)
    pd_pool_meta = AccountMeta(pd_pool_pubkey, False, True)
    stake_program_meta = AccountMeta(ingl_constants.STAKE_PROGRAM_ID, False, False)
    vote_account_meta = AccountMeta(expected_vote_pubkey, False, True)
    stake_history_meta = AccountMeta(STAKE_HISTORY, False, False)
    stake_config_meta = AccountMeta(
        ingl_constants.STAKE_CONFIG_PROGRAM_ID, False, False
    )
    config_account_meta = AccountMeta(config_account_pubkey, False, False)

    accounts = [
        payer_account_meta,
        t_stake_meta,
        pd_pool_meta,
        general_account_meta,
        sysvar_clock_meta,
        rent_account_meta,
        stake_account_meta,
        t_withdraw_meta,
        vote_account_meta,
        stake_history_meta,
        stake_config_meta,
        config_account_meta,
        sys_program_meta,
        stake_program_meta,
        sys_program_meta,
        sys_program_meta,
        stake_program_meta,
        stake_program_meta,
    ]
    # print(accounts)
    data = InstructionEnum.build(
        InstructionEnum.enum.InitRebalance(
            log_level=log_level,
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def finalize_rebalance(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    expected_stake_key, _expected_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_stake_key, _t_stake_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_STAKE_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    t_withdraw_key, _t_withdraw_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.T_WITHDRAW_KEY, "UTF-8")], get_program_id()
    )
    pd_pool_pubkey, _pd_pool_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.PD_POOL_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )

    data = await client.get_account_info(config_account_pubkey)
    validator_id = Pubkey(ValidatorConfig.parse(data.value.data).validator_id)
    print(f"Validator_Id: {validator_id}")

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    sysvar_clock_meta = AccountMeta(solders.sysvar.CLOCK, False, False)
    validator_meta = AccountMeta(validator_id, False, True)
    stake_account_meta = AccountMeta(expected_stake_key, False, True)
    t_stake_meta = AccountMeta(t_stake_key, False, True)
    t_withdraw_meta = AccountMeta(t_withdraw_key, False, True)
    pd_pool_meta = AccountMeta(pd_pool_pubkey, False, True)
    stake_program_meta = AccountMeta(ingl_constants.STAKE_PROGRAM_ID, False, False)
    sysvar_stake_history_meta = AccountMeta(solders.sysvar.STAKE_HISTORY, False, False)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    config_account_meta = AccountMeta(config_account_pubkey, False, False)
    system_program_meta = AccountMeta(solders.system_program.ID, False, False)

    accounts = [
        payer_account_meta,
        validator_meta,
        t_stake_meta,
        pd_pool_meta,
        general_account_meta,
        sysvar_clock_meta,
        stake_account_meta,
        t_withdraw_meta,
        sysvar_stake_history_meta,
        config_account_meta,
        stake_program_meta,
        system_program_meta,
    ]

    data = InstructionEnum.build(
        InstructionEnum.enum.FinalizeRebalance(
            log_level=log_level,
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def process_rewards(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    ingl_team_account_pubkey = Pubkey.from_string(
        "Team111111111111111111111111111111111111111"
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )


async def process_rewards(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    ingl_team_account_pubkey = Pubkey.from_string(
        "Team111111111111111111111111111111111111111"
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    config_data = ValidatorConfig.parse(
        (await client.get_account_info(config_account_pubkey)).value.data
    )
    vote_account_key = Pubkey(config_data.vote_account)
    validator_id = Pubkey(config_data.validator_id)
    print(f"Validator_Id: {validator_id}")

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    validator_meta = AccountMeta(validator_id, False, True)
    vote_account_meta = AccountMeta(vote_account_key, False, True)
    sys_program_meta = AccountMeta(system_program.ID, False, False)
    vote_program_meta = AccountMeta(ingl_constants.VOTE_PROGRAM_ID, False, False)
    mint_authority_meta = AccountMeta(ingl_team_account_pubkey, False, True)
    authorized_withdrawer_meta = AccountMeta(authorized_withdrawer_key, False, True)
    config_account_meta = AccountMeta(config_account_pubkey, False, False)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)

    accounts = [
        payer_account_meta,
        validator_meta,
        vote_account_meta,
        authorized_withdrawer_meta,
        config_account_meta,
        general_account_meta,
        mint_authority_meta,
        vote_program_meta,
        sys_program_meta,
        sys_program_meta,
        sys_program_meta,
    ]
    # print(accounts)
    config_data = InstructionEnum.build(
        InstructionEnum.enum.ProcessRewards(
            log_level=log_level,
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=config_data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def nft_withdraw(
    payer_keypair: KeypairInput,
    mints: List[Pubkey],
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [
            bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8"),
            bytes(vote_account_id.pubkey),
        ],
        get_program_id(),
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )


async def nft_withdraw(
    payer_keypair: KeypairInput,
    mints: List[Pubkey],
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [
            bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8"),
            bytes(vote_account_id.pubkey),
        ],
        get_program_id(),
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )

    config_data = ValidatorConfig.parse(
        (await client.get_account_info(config_account_pubkey)).value.data
    )
    vote_account_id = Pubkey(config_data.vote_account)

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    vote_account_meta = AccountMeta(vote_account_id, False, True)
    sys_program_meta = AccountMeta(system_program.ID, False, False)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    authorized_withdrawer_meta = AccountMeta(authorized_withdrawer_key, False, True)

    accounts = [
        payer_account_meta,
        vote_account_meta,
        general_account_meta,
        authorized_withdrawer_meta,
        sys_program_meta,
    ]

    for mint_pubkey in mints:
        mint_associated_account_pubkey = (
            assoc_instructions.get_associated_token_address(
                payer_keypair.pubkey, mint_pubkey.pubkey
            )
        )
        accounts.append(AccountMeta(mint_associated_account_pubkey, False, False))
        accounts.append(AccountMeta(mint_pubkey.pubkey, False, False))
        gem_account_pubkey, _gem_account_bump = Pubkey.find_program_address(
            [
                bytes(ingl_constants.GEM_ACCOUNT_CONST, "UTF-8"),
                bytes(mint_pubkey.pubkey),
            ],
            get_program_id(),
        )
        accounts.append(AccountMeta(gem_account_pubkey, False, True))

    accounts.append(sys_program_meta)
    # print(accounts)
    data = InstructionEnum.build(
        InstructionEnum.enum.NFTWithdraw(log_level=log_level, cnt=len(mints))
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def init_governance(
    payer_keypair: KeypairInput,
    mint: PubkeyInput,
    client: AsyncClient,
    title: str,
    description: str,
    governance_type: Optional[GovernanceType.enum] = None,
    config_account_type: Optional[ConfigAccountType.enum] = None,
    vote_account_governance: Optional[VoteAccountGovernance.enum] = None,
    log_level: int = 0,
) -> str:
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    config_account_pubkey, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    nft_account_data_pubkey, _nft_account_data_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint.pubkey)],
        get_program_id(),
    )
    associated_account_key = assoc_instructions.get_associated_token_address(
        payer_keypair.pubkey, mint.pubkey
    )

    data = await client.get_account_info(general_account_pubkey)
    proposal_numeration = GeneralData.parse(data.value.data).proposal_numeration
    print(f"proposal_numeration: {proposal_numeration}")

    config_data = ValidatorConfig.parse(
        (await client.get_account_info(config_account_pubkey)).value.data
    )
    vote_account_key = Pubkey(config_data.vote_account)

    proposal_pubkey, _proposal_bump = Pubkey.find_program_address(
        [
            bytes(ingl_constants.INGL_PROPOSAL_KEY, "UTF-8"),
            (proposal_numeration).to_bytes(4, "big"),
        ],
        get_program_id(),
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    vote_account_meta = AccountMeta(
        pubkey=vote_account_key, is_signer=False, is_writable=True
    )
    proposal_account_meta = AccountMeta(
        pubkey=proposal_pubkey, is_signer=False, is_writable=True
    )
    general_account_meta = AccountMeta(
        pubkey=general_account_pubkey, is_signer=False, is_writable=True
    )
    mint_account_meta = AccountMeta(
        pubkey=mint.pubkey, is_signer=False, is_writable=False
    )
    associated_account_meta = AccountMeta(
        pubkey=associated_account_key, is_signer=False, is_writable=False
    )
    nft_account_data_meta = AccountMeta(
        pubkey=nft_account_data_pubkey, is_signer=False, is_writable=False
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_pubkey, is_signer=False, is_writable=False
    )

    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    accounts = [
        payer_account_meta,
        vote_account_meta,
        proposal_account_meta,
        general_account_meta,
        mint_account_meta,
        associated_account_meta,
        nft_account_data_meta,
        config_account_meta,
    ]

    instruction_data = build_instruction(
        InstructionEnum.enum.InitGovernance(),
        title=title,
        description=description,
        governance_type=governance_type,
        config_account_type=config_account_type,
        vote_account_governance=vote_account_governance,
        log_level=log_level,
    )
    print("instruction_data: ", instruction_data)
    if instruction_data[1] == 1:
        buffer_account_key = Pubkey(instruction_data[2:34])
        buffer_account_meta = AccountMeta(
            pubkey=buffer_account_key, is_signer=False, is_writable=True
        )
        print("buffer_account_key: ", buffer_account_key)
        accounts.append(buffer_account_meta)

    accounts += [
        system_program_meta,
    ]

    # print(accounts)
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def vote_governance(
    payer_keypair: KeypairInput,
    vote: Bool,
    proposal_numeration: int,
    mints: List[Pubkey],
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    proposal_pubkey, _proposal_account_bump = Pubkey.find_program_address(
        [
            bytes(ingl_constants.INGL_PROPOSAL_KEY, "UTF-8"),
            (proposal_numeration).to_bytes(4, "big"),
        ],
        get_program_id(),
    )

    print(
        f"Proposal_Account: {proposal_pubkey}, Vote: {'Approve' if vote else 'Dissaprove'} "
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    proposal_account_meta = AccountMeta(
        pubkey=proposal_pubkey, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    accounts = [
        payer_account_meta,
        proposal_account_meta,
    ]

    for mint in mints:
        associated_account_key = assoc_instructions.get_associated_token_address(
            payer_keypair.pubkey, mint
        )
        nft_account_pubkey, _nft_account_bump = Pubkey.find_program_address(
            [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint)],
            get_program_id(),
        )

        nft_account_meta = AccountMeta(
            pubkey=nft_account_pubkey, is_signer=False, is_writable=True
        )
        mint_account_meta = AccountMeta(pubkey=mint, is_signer=False, is_writable=False)
        associated_account_meta = AccountMeta(
            pubkey=associated_account_key, is_signer=False, is_writable=False
        )
        accounts += [
            nft_account_meta,
            mint_account_meta,
            associated_account_meta,
        ]
    accounts.append(system_program_meta)

    # print(accounts)
    instruction_data = build_instruction(
        InstructionEnum.enum.VoteGovernance(
            log_level=log_level,
            numeration=proposal_numeration,
            vote=vote,
            cnt=len(mints),
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def finalize_governance(
    payer_keypair: KeypairInput,
    proposal_numeration: int,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    proposal_account_key, _proposal_account_bump = Pubkey.find_program_address(
        [
            bytes(ingl_constants.INGL_PROPOSAL_KEY, "UTF-8"),
            (proposal_numeration).to_bytes(4, "big"),
        ],
        get_program_id(),
    )
    config_account_key, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_key, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    print(f"Proposal_Account: {proposal_account_key}")

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    sysvar_rent_account_meta = AccountMeta(
        pubkey=RENT, is_signer=False, is_writable=False
    )
    sysvar_clock_account_meta = AccountMeta(
        pubkey=CLOCK, is_signer=False, is_writable=False
    )
    proposal_account_meta = AccountMeta(
        pubkey=proposal_account_key, is_signer=False, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    general_account_meta = AccountMeta(
        pubkey=general_account_key, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        sysvar_rent_account_meta,
        sysvar_clock_account_meta,
        proposal_account_meta,
        config_account_meta,
        general_account_meta,
    ]

    # print(accounts)
    instruction_data = build_instruction(
        InstructionEnum.enum.FinalizeGovernance(
            log_level=log_level, numeration=proposal_numeration
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def execute_governance(
    payer_keypair: KeypairInput,
    proposal_numeration: int,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    proposal_account_key, _proposal_account_bump = Pubkey.find_program_address(
        [
            bytes(ingl_constants.INGL_PROPOSAL_KEY, "UTF-8"),
            (proposal_numeration).to_bytes(4, "big"),
        ],
        get_program_id(),
    )
    config_account_key, _config_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_key, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    print(f"Proposal_Account: {proposal_account_key}")

    proposal_data = await client.get_account_info(proposal_account_key)
    proposal_data = proposal_data.value.data

    # payer is the validator ID.
    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    sysvar_rent_account_meta = AccountMeta(
        pubkey=RENT, is_signer=False, is_writable=False
    )
    sysvar_clock_account_meta = AccountMeta(
        pubkey=CLOCK, is_signer=False, is_writable=False
    )
    proposal_account_meta = AccountMeta(
        pubkey=proposal_account_key, is_signer=False, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    general_account_meta = AccountMeta(
        pubkey=general_account_key, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        sysvar_clock_account_meta,
        proposal_account_meta,
        config_account_meta,
        general_account_meta,
    ]

    loc = 4 + 4 + 1
    if proposal_data[loc] == 1:
        loc += 1 + 4
    else:
        loc += 1

    if proposal_data[loc] == 1:
        loc += 1 + 1
    else:
        loc += 1

    loc += 1

    title_len = int.from_bytes(proposal_data[loc : loc + 4], "big")
    loc += 4
    loc += title_len

    desc_len = int.from_bytes(proposal_data[loc : loc + 4], "big")
    loc += 4
    loc += desc_len

    num_votes = int.from_bytes(proposal_data[loc : loc + 4], "big")
    loc += 4
    loc += num_votes * 5
    governance_type = proposal_data[loc]
    loc += 1
    if governance_type == 0:
        pass
    elif governance_type == 1:
        buffer_account_key = Pubkey(proposal_data[loc : loc + 32])
        programdata_key, _programdata_bump = Pubkey.find_program_address(
            [bytes(get_program_id())], ingl_constants.BPF_LOADER_UPGRADEABLE
        )
        upgrade_authority_key, _upgrade_authority_bump = Pubkey.find_program_address(
            [
                bytes(ingl_constants.INGL_PROGRAM_AUTHORITY_KEY, "UTF-8"),
                bytes(get_program_id()),
            ],
            get_program_id(),
        )

        upgraded_program_meta = AccountMeta(
            pubkey=get_program_id(), is_signer=False, is_writable=False
        )
        buffer_account_meta = AccountMeta(
            pubkey=buffer_account_key, is_signer=False, is_writable=True
        )
        programdata_account_meta = AccountMeta(
            pubkey=programdata_key, is_signer=False, is_writable=True
        )
        upgrade_authority_account_meta = AccountMeta(
            pubkey=upgrade_authority_key, is_signer=False, is_writable=True
        )

        accounts += [
            upgraded_program_meta,
            buffer_account_meta,
            payer_account_meta,
            programdata_account_meta,
            upgrade_authority_account_meta,
            sysvar_rent_account_meta,
            sysvar_clock_account_meta,
        ]
    elif governance_type == 2:
        vote_account_governance_type = proposal_data[loc]
        loc += 1
        vote_account_key, _vote_account_bump = Pubkey.find_program_address(
            [
                bytes(ingl_constants.INGL_VOTE_KEY, "UTF-8"),
            ],
            get_program_id(),
        )
        (
            authorized_withdrawer_key,
            _authorized_withdrawer_bump,
        ) = Pubkey.find_program_address(
            [
                bytes(ingl_constants.INGL_VOTE_AUTHORIZED_WITHDRAWER_KEY, "UTF-8"),
            ],
            get_program_id(),
        )

        vote_account_meta = AccountMeta(
            pubkey=vote_account_key, is_signer=False, is_writable=True
        )
        authorized_withdrawer_account_meta = AccountMeta(
            pubkey=authorized_withdrawer_key, is_signer=False, is_writable=False
        )

        accounts += [
            vote_account_meta,
            authorized_withdrawer_account_meta,
        ]

        if vote_account_governance_type == 0:
            accounts.append(sysvar_clock_account_meta)
            accounts.append(payer_account_meta)

    # print(accounts)
    instruction_data = build_instruction(
        InstructionEnum.enum.ExecuteGovernance(
            log_level=log_level, numeration=proposal_numeration
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def reset_uris(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )
    # payer is the validator id address.
    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    uris_account_key, uris_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_URIS_ACCOUNT_KEY, "UTF-8")], get_program_id()
    )

    uris_account_meta = AccountMeta(
        pubkey=uris_account_key, is_signer=False, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.SYSTEM_PROGRAM_ID, is_signer=False, is_writable=False
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        uris_account_meta,
        system_program_meta,
    ]

    instruction_data = build_instruction(
        InstructionEnum.enum.ResetUris(log_level=log_level)
    )
    transaction = Transaction()
    transaction.add(
        Instruction(
            accounts=accounts, program_id=get_program_id(), data=instruction_data
        )
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


def upload_uris(
    payer_keypair: KeypairInput,
    upload_authority: KeypairInput,
    uris: List[str],
    rarity: int,
    client: Client,
    log_level: int = 0,
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    uris_account_key, _config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.URIS_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )
    uris_account_meta = AccountMeta(
        pubkey=uris_account_key, is_signer=False, is_writable=True
    )
    upload_authority_account_meta = AccountMeta(
        pubkey=upload_authority.pubkey, is_signer=True, is_writable=False
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        uris_account_meta,
        upload_authority_account_meta,
        system_program_meta,
    ]

    t_dets = None
    try:
        instruction_data = build_instruction(
            InstructionEnum.enum.UploadUris(
                uris=uris, rarity=rarity, log_level=log_level
            )
        )
        transaction = Transaction()
        transaction.add(
            ComputeBudgetInstruction().set_compute_unit_limit(
                1_000_000, payer_keypair.pubkey
            )
        )
        transaction.add(
            Instruction(
                accounts=accounts, program_id=get_program_id(), data=instruction_data
            )
        )
        t_dets = client.send_transaction(transaction, payer_keypair.keypair)
        return t_dets
    except Exception as e:
        print(t_dets, e)
        raise e


async def reset_uris(
    payer_keypair: KeypairInput, client: AsyncClient, log_level: int = 0
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    uris_account_key, _config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.URIS_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )
    uris_account_meta = AccountMeta(
        pubkey=uris_account_key, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        uris_account_meta,
        system_program_meta,
    ]

    t_dets = None
    try:
        instruction_data = build_instruction(
            InstructionEnum.enum.ResetUris(log_level=log_level)
        )
        transaction = Transaction()
        transaction.add(
            ComputeBudgetInstruction().set_compute_unit_limit(
                1_000_000, payer_keypair.pubkey
            )
        )
        transaction.add(
            Instruction(
                accounts=accounts, program_id=get_program_id(), data=instruction_data
            )
        )
        t_dets = await client.send_transaction(transaction, payer_keypair.keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        print(t_dets, e)
        raise e


async def init_registry(
    payer_keypair: KeypairInput,
    client: AsyncClient,
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [b"config"], ingl_constants.REGISTRY_PROGRAM
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        system_program_meta,
    ]

    t_dets = None
    try:
        instruction_data = RegistryEnum.build(RegistryEnum.enum.InitConfig())
        transaction = Transaction()
        transaction.add(
            Instruction(
                accounts=accounts,
                program_id=ingl_constants.REGISTRY_PROGRAM,
                data=instruction_data,
            )
        )
        t_dets = await client.send_transaction(transaction, payer_keypair.keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        print(t_dets, e)
        raise e


async def reset_registry(
    payer_keypair: KeypairInput,
    client: AsyncClient,
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [b"config"], ingl_constants.REGISTRY_PROGRAM
    )

    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )
    storage_account_key, _config_account_bump = Pubkey.find_program_address(
        [b"storage"],
        ingl_constants.REGISTRY_PROGRAM,
    )
    name_storage_key, _name_storage_bump = Pubkey.find_program_address(
        [b"name_storage"],
        ingl_constants.REGISTRY_PROGRAM,
    )
    marketplace_storage_key, _marketplace_storage_bump = Pubkey.find_program_address(
        [b"marketplace_storage"],
        ingl_constants.REGISTRY_PROGRAM,
    )

    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )

    name_storage_meta = AccountMeta(
        pubkey=name_storage_key, is_signer=False, is_writable=True
    )

    marketplace_storage_meta = AccountMeta(
        pubkey=marketplace_storage_key, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        storage_account_meta,
        name_storage_meta,
        marketplace_storage_meta,
        system_program_meta,
    ]

    t_dets = None
    try:
        instruction_data = RegistryEnum.build(RegistryEnum.enum.Reset())
        transaction = Transaction()
        transaction.add(
            Instruction(
                accounts=accounts,
                program_id=ingl_constants.REGISTRY_PROGRAM,
                data=instruction_data,
            )
        )
        t_dets = await client.send_transaction(transaction, payer_keypair.keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        print(t_dets, e)
        raise e


async def register_program(
    payer_keypair: KeypairInput,
    program_key: Pubkey,
    client: AsyncClient,
) -> str:
    config_account_key, _config_bump = Pubkey.find_program_address(
        [b"config"], ingl_constants.REGISTRY_PROGRAM
    )
    payer_account_meta = AccountMeta(
        pubkey=payer_keypair.pubkey, is_signer=True, is_writable=True
    )
    config_account_meta = AccountMeta(
        pubkey=config_account_key, is_signer=False, is_writable=True
    )
    registered_program_meta = AccountMeta(
        pubkey=program_key, is_signer=False, is_writable=False
    )
    team_account_meta = AccountMeta(
        pubkey=ingl_constants.TEAM_ADDRESS, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    config_data = await client.get_account_info(config_account_key)
    config_data = RegistryConfig.parse(config_data.value.data)
    storage_account_key, _config_account_bump = Pubkey.find_program_address(
        [b"storage", (config_data.validator_numeration // 625).to_bytes(4, "big")],
        ingl_constants.REGISTRY_PROGRAM,
    )
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )

    accounts = [
        payer_account_meta,
        config_account_meta,
        registered_program_meta,
        team_account_meta,
        storage_account_meta,
        system_program_meta,
    ]

    t_dets = None
    try:
        instruction_data = RegistryEnum.build(RegistryEnum.enum.AddProgram(name=name))
        transaction = Transaction()
        transaction.add(
            Instruction(
                accounts=accounts,
                program_id=ingl_constants.REGISTRY_PROGRAM,
                data=instruction_data,
            )
        )
        t_dets = await client.send_transaction(transaction, payer_keypair.keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        print(t_dets, e)
        raise e


async def fractionalize_existing(
    payer_keypair: KeypairInput,
    authorized_withdrawer: KeypairInput,
    upgrade_authority: KeypairInput,
    vote_account_pubkey: PubkeyInput,
    validator_pubkey: PubkeyInput,
    init_commission: int,
    max_primary_stake: int,
    nft_holders_share: int,
    initial_redemption_fee: int,
    is_validator_id_switchable: bool,
    unit_backing: int,
    redemption_fee_duration: int,
    proposal_quorum: int,
    creator_royalties: int,
    governance_expiration_time: int,
    rarities: List[int],
    rarity_names: List[str],
    twitter_handle: str,
    discord_invite: str,
    validator_name: str,
    collection_uri: str,
    website: str,
    default_uri: str,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    mint_pubkey, _mint_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_NFT_COLLECTION_KEY, "UTF-8")], get_program_id()
    )
    mint_authority_pubkey, _mint_authority_pubkey_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_MINT_AUTHORITY_KEY, "UTF-8")], get_program_id()
    )
    (
        collection_holder_pubkey,
        _collection_holder_pubkey_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.COLLECTION_HOLDER_KEY, "UTF-8")], get_program_id()
    )
    mint_associated_account_pubkey = assoc_instructions.get_associated_token_address(
        collection_holder_pubkey, mint_pubkey
    )
    metaplex_program_id = Pubkey.from_string(
        "metaqbxxUerdq28cj1RbAWkYQm3ybzjb6a8bt518x1s"
    )
    metadata_pda, _metadata_pda_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(mint_pubkey)],
        metaplex_program_id,
    )
    master_edition_pda, _master_edition_bump = Pubkey.find_program_address(
        [b"metadata", bytes(metaplex_program_id), bytes(mint_pubkey), b"edition"],
        metaplex_program_id,
    )
    ingl_config_pubkey, _ingl_config_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_CONFIG_SEED, "UTF-8")], get_program_id()
    )
    general_account_pubkey, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )
    uris_account_pubkey, _uris_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.URIS_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    (
        registry_program_config_key,
        _registry_program_config_bump,
    ) = Pubkey.find_program_address([b"config"], ingl_constants.REGISTRY_PROGRAM_ID)

    registry_config_account = await client.get_account_info(registry_program_config_key)
    registry_config_data = RegistryConfig.parse(registry_config_account.value.data)
    storage_numeration = registry_config_data.validator_numeration // 625
    name_storage_numeration = registry_config_data.validator_numeration // 1666
    storage_key, _storage_bump = Pubkey.find_program_address(
        [b"storage", storage_numeration.to_bytes(4, "big")],
        ingl_constants.REGISTRY_PROGRAM_ID,
    )

    (
        pda_authorized_withdrawer,
        _pda_authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.INGL_AUTHORIZED_WITHDRAWER_KEY, "UTF-8")],
        get_program_id(),
    )

    programdata_key, _programdata_bump = Pubkey.find_program_address(
        [bytes(get_program_id())], ingl_constants.BPF_LOADER_UPGRADEABLE
    )

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    collection_holder_meta = AccountMeta(collection_holder_pubkey, False, True)
    mint_account_meta = AccountMeta(mint_pubkey, False, True)
    mint_authority_meta = AccountMeta(mint_authority_pubkey, False, False)
    mint_associated_meta = AccountMeta(mint_associated_account_pubkey, False, True)
    spl_program_meta = AccountMeta(spl_constants.TOKEN_PROGRAM_ID, False, False)
    sysvar_rent_account_meta = AccountMeta(solders.sysvar.RENT, False, False)
    system_program_meta = AccountMeta(system_program.ID, False, False)
    token_metadata_meta = AccountMeta(metadata_pda, False, True)
    metadata_program_id = AccountMeta(metaplex_program_id, False, False)
    associated_program_meta = AccountMeta(
        spl_constants.ASSOCIATED_TOKEN_PROGRAM_ID, False, False
    )
    edition_meta = AccountMeta(master_edition_pda, False, True)
    ingl_config_meta = AccountMeta(ingl_config_pubkey, False, True)
    general_account_meta = AccountMeta(general_account_pubkey, False, True)
    uris_account_meta = AccountMeta(uris_account_pubkey, False, True)
    validator_account_meta = AccountMeta(validator_pubkey.pubkey, False, True)
    registry_program_config_meta = AccountMeta(registry_program_config_key, False, True)
    program_meta = AccountMeta(get_program_id(), False, False)
    team_account_meta = AccountMeta(ingl_constants.TEAM_ACCOUNT_KEY, False, True)
    storage_account_meta = AccountMeta(storage_key, False, True)
    registry_program_meta = AccountMeta(
        ingl_constants.REGISTRY_PROGRAM_ID, False, False
    )
    current_authorized_withdrawer_meta = AccountMeta(
        authorized_withdrawer.pubkey, True, False
    )
    pda_authorized_withdrawer_meta = AccountMeta(
        pda_authorized_withdrawer, False, False
    )
    vote_account_meta = AccountMeta(vote_account_pubkey.pubkey, False, True)
    sysvar_clock_meta = AccountMeta(solders.sysvar.CLOCK, False, False)
    this_program_data_meta = AccountMeta(
        pubkey=programdata_key, is_signer=False, is_writable=False
    )
    current_upgrade_authority_meta = AccountMeta(upgrade_authority, True, False)

    accounts = [
        payer_account_meta,
        ingl_config_meta,
        general_account_meta,
        uris_account_meta,
        sysvar_rent_account_meta,
        validator_account_meta,
        collection_holder_meta,
        mint_account_meta,
        mint_authority_meta,
        mint_associated_meta,
        token_metadata_meta,
        edition_meta,
        spl_program_meta,
        system_program_meta,
        this_program_data_meta,
        current_upgrade_authority_meta,
        current_authorized_withdrawer_meta,
        pda_authorized_withdrawer_meta,
        vote_account_meta,
        sysvar_clock_meta,
        program_meta,
        team_account_meta,
        storage_account_meta,
        system_program_meta,
        associated_program_meta,
        spl_program_meta,
        metadata_program_id,
        registry_program_meta,
    ]
    # print(accounts)
    data = build_instruction(
        InstructionEnum.enum.FractionalizeExisting(
            init_commission=init_commission,
            max_primary_stake=max_primary_stake,
            nft_holders_share=nft_holders_share,
            initial_redemption_fee=initial_redemption_fee,
            is_validator_id_switchable=is_validator_id_switchable,
            unit_backing=unit_backing,
            redemption_fee_duration=redemption_fee_duration,
            proposal_quorum=proposal_quorum,
            creator_royalties=creator_royalties,
            governance_expiration_time=governance_expiration_time,
            name_storage_numeration=name_storage_numeration,
            rarities=rarities,
            rarity_names=rarity_names,
            twitter_handle=twitter_handle,
            discord_invite=discord_invite,
            validator_name=validator_name,
            collection_uri=collection_uri,
            website=website,
            default_uri=default_uri,
            log_level=log_level,
        )
    )
    transaction = Transaction()
    # print(data)
    transaction.add(
        ComputeBudgetInstruction().set_compute_unit_limit(400_000, payer_keypair.pubkey)
    )
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )

    try:
        t_dets = await sign_and_send_tx(
            transaction, client, payer_keypair, authorized_withdrawer
        )
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"Error: {e}"


async def inject_testing_data(
    payer_keypair: KeypairInput,
    mints: List[Pubkey],
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    (
        authorized_withdrawer_key,
        _authorized_withdrawer_bump,
    ) = Pubkey.find_program_address(
        [bytes(ingl_constants.AUTHORIZED_WITHDRAWER_KEY, "UTF-8")], get_program_id()
    )
    general_account_key, _general_account_bump = Pubkey.find_program_address(
        [bytes(ingl_constants.GENERAL_ACCOUNT_SEED, "UTF-8")], get_program_id()
    )

    payer_account_meta = AccountMeta(payer_keypair.pubkey, True, True)
    general_account_meta = AccountMeta(general_account_key, False, True)
    sys_program_meta = AccountMeta(system_program.ID, False, False)
    authorized_withdrawer_meta = AccountMeta(authorized_withdrawer_key, False, True)

    accounts = [
        payer_account_meta,
        general_account_meta,
        authorized_withdrawer_meta,
    ]

    for mint_pubkey in mints:
        accounts.append(AccountMeta(mint_pubkey, False, False))
        nft_account_pubkey, _nft_account_bump = Pubkey.find_program_address(
            [bytes(ingl_constants.NFT_ACCOUNT_CONST, "UTF-8"), bytes(mint_pubkey)],
            get_program_id(),
        )
        accounts.append(AccountMeta(nft_account_pubkey, False, True))

    accounts.append(sys_program_meta)
    # print(accounts)
    data = InstructionEnum.build(
        InstructionEnum.enum.InjectTestingData(
            log_level=log_level, num_mints=len(mints)
        )
    )
    transaction = Transaction()
    transaction.add(
        Instruction(accounts=accounts, program_id=get_program_id(), data=data)
    )
    try:
        t_dets = await sign_and_send_tx(transaction, client, payer_keypair)
        await client.confirm_transaction(
            tx_sig=t_dets.value,
            commitment="finalized",
            sleep_seconds=0.4,
            last_valid_block_height=None,
        )
        return f"Transaction Id: [link=https://explorer.solana.com/tx/{str(t_dets.value)+get_explorer_suffix(get_network())}]{str(t_dets.value)}[/link]"
    except Exception as e:
        return f"[warning]Error: {e}[/warning]"


# Market Processes are below:


async def process_list_validator(
    authorized_withdrawer: KeypairInput,
    current_upgrade_authority: KeypairInput,
    vote_account: PubkeyInput,
    authorized_withdrawer_cost: int,
    mediatable_date: int,
    secondary_items: List[dict],
    description: str,
    validator_name: str,
    validator_logo_url: str,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    pda_authorized_withdrawer_key, _pda_aw_bump = Pubkey.find_program_address(
        [ingl_constants.PDA_AUTHORIZED_WITHDRAWER_SEED], get_market_program_id()
    )
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )
    this_program_data_key, _this_program_data_bump = Pubkey.find_program_address(
        [bytes(get_market_program_id())], ingl_constants.BPF_LOADER_UPGRADEABLE
    )
    (
        pda_upgrade_authority_key,
        _pda_upgrade_authority_bump,
    ) = Pubkey.find_program_address(
        [ingl_constants.PDA_UPGRADE_AUTHORITY_SEED], get_market_program_id()
    )
    registry_storage_key, _registry_storage_bump = Pubkey.find_program_address(
        [ingl_constants.REGISTRY_STORAGE_SEED], ingl_constants.REGISTRY_PROGRAM
    )

    # Forming AccountMetas
    authorized_withdrawer_meta = AccountMeta(
        pubkey=authorized_withdrawer.pubkey, is_signer=True, is_writable=True
    )
    vote_account_meta = AccountMeta(
        pubkey=vote_account.pubkey, is_signer=False, is_writable=True
    )
    pda_authorized_withdrawer_meta = AccountMeta(
        pubkey=pda_authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    this_program_meta = AccountMeta(
        pubkey=get_market_program_id(), is_signer=False, is_writable=True
    )
    this_program_data_meta = AccountMeta(
        pubkey=this_program_data_key, is_signer=False, is_writable=True
    )
    current_upgrade_authority_meta = AccountMeta(
        pubkey=current_upgrade_authority.pubkey, is_signer=True, is_writable=False
    )
    pda_upgrade_authority_meta = AccountMeta(
        pubkey=pda_upgrade_authority_key, is_signer=False, is_writable=False
    )
    sysvar_clock_meta = AccountMeta(
        pubkey=solders.sysvar.CLOCK, is_signer=False, is_writable=False
    )
    team_account_meta = AccountMeta(
        pubkey=ingl_constants.TEAM_ADDRESS, is_signer=False, is_writable=True
    )
    registry_storage_meta = AccountMeta(
        pubkey=registry_storage_key, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )
    rent_sysvar_meta = AccountMeta(
        pubkey=solders.sysvar.RENT, is_signer=False, is_writable=False
    )
    vote_program_meta = AccountMeta(
        pubkey=ingl_constants.VOTE_PROGRAM_ID, is_signer=False, is_writable=False
    )
    bpf_loader_upgradeable_meta = AccountMeta(
        pubkey=ingl_constants.BPF_LOADER_UPGRADEABLE, is_signer=False, is_writable=False
    )
    register_program_meta = AccountMeta(
        pubkey=ingl_constants.REGISTRY_PROGRAM, is_signer=False, is_writable=False
    )

    account_metas = [
        authorized_withdrawer_meta,
        vote_account_meta,
        pda_authorized_withdrawer_meta,
        storage_account_meta,
        this_program_meta,
        this_program_data_meta,
        current_upgrade_authority_meta,
        pda_upgrade_authority_meta,
        sysvar_clock_meta,
        team_account_meta,
        registry_storage_meta,
        system_program_meta,
        rent_sysvar_meta,
        # programs listed below
        vote_program_meta,
        bpf_loader_upgradeable_meta,
        register_program_meta,
    ]
    # print(account_metas)

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.List(
                log_level=log_level,
                authorized_withdrawer_cost=authorized_withdrawer_cost,
                mediatable_date=mediatable_date,
                secondary_items=secondary_items,
                description=description,
                validator_name=validator_name,
                validator_logo_url=validator_logo_url,
            ),
            account_metas,
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(
        transaction, client, authorized_withdrawer, current_upgrade_authority
    )


async def process_delist_validator(
    authorized_withdrawer: KeypairInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    pda_authorized_withdrawer_key, _pda_aw_bump = Pubkey.find_program_address(
        [ingl_constants.PDA_AUTHORIZED_WITHDRAWER_SEED], get_market_program_id()
    )
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )
    this_program_data_key, _this_program_data_bump = Pubkey.find_program_address(
        [bytes(get_market_program_id())], ingl_constants.BPF_LOADER_UPGRADEABLE
    )
    (
        pda_upgrade_authority_key,
        _pda_upgrade_authority_bump,
    ) = Pubkey.find_program_address(
        [ingl_constants.PDA_UPGRADE_AUTHORITY_SEED], get_market_program_id()
    )

    storage_account_data = Storage.parse(
        (await client.get_account_info(storage_account_key)).value.data
    )
    vote_account_key = Pubkey(storage_account_data.vote_account)

    # Forming AccountMetas
    authorized_withdrawer_meta = AccountMeta(
        pubkey=authorized_withdrawer.pubkey, is_signer=True, is_writable=True
    )
    vote_account_meta = AccountMeta(
        pubkey=vote_account_key, is_signer=False, is_writable=True
    )
    pda_authorized_withdrawer_meta = AccountMeta(
        pubkey=pda_authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    this_program_meta = AccountMeta(
        pubkey=get_market_program_id(), is_signer=False, is_writable=True
    )
    this_program_data_meta = AccountMeta(
        pubkey=this_program_data_key, is_signer=False, is_writable=True
    )
    pda_upgrade_authority_meta = AccountMeta(
        pubkey=pda_upgrade_authority_key, is_signer=False, is_writable=False
    )
    sysvar_clock_meta = AccountMeta(
        pubkey=solders.sysvar.CLOCK, is_signer=False, is_writable=False
    )
    vote_program_meta = AccountMeta(
        pubkey=ingl_constants.VOTE_PROGRAM_ID, is_signer=False, is_writable=False
    )
    bpf_loader_upgradeable_meta = AccountMeta(
        pubkey=ingl_constants.BPF_LOADER_UPGRADEABLE, is_signer=False, is_writable=False
    )

    account_metas = [
        authorized_withdrawer_meta,
        vote_account_meta,
        pda_authorized_withdrawer_meta,
        storage_account_meta,
        this_program_meta,
        this_program_data_meta,
        pda_upgrade_authority_meta,
        sysvar_clock_meta,
        # programs listed below
        vote_program_meta,
        bpf_loader_upgradeable_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.Delist(log_level=log_level), account_metas
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(
        transaction, client, authorized_withdrawer
    )


async def process_buy_validator(
    payer: KeypairInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    pda_authorized_withdrawer_key, _pda_aw_bump = Pubkey.find_program_address(
        [ingl_constants.PDA_AUTHORIZED_WITHDRAWER_SEED], get_market_program_id()
    )
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )
    escrow_account_key, _escrow_bump = Pubkey.find_program_address(
        [ingl_constants.ESCROW_ACCOUNT_SEED], get_market_program_id()
    )

    storage_account_data = Storage.parse(
        (await client.get_account_info(storage_account_key)).value.data
    )
    vote_account_key = Pubkey(storage_account_data.vote_account)
    registered_authorized_withdrawer_key = Pubkey(
        storage_account_data.authorized_withdrawer
    )

    # Forming AccountMetas
    payer_meta = AccountMeta(pubkey=payer.pubkey, is_signer=True, is_writable=True)
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    registered_authorized_withdrawer_meta = AccountMeta(
        pubkey=registered_authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    vote_account_meta = AccountMeta(
        pubkey=vote_account_key, is_signer=False, is_writable=True
    )
    sysvar_clock_meta = AccountMeta(
        pubkey=solders.sysvar.CLOCK, is_signer=False, is_writable=False
    )
    pda_authorized_withdrawer_meta = AccountMeta(
        pubkey=pda_authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    escrow_account_meta = AccountMeta(
        pubkey=escrow_account_key, is_signer=False, is_writable=True
    )
    team_account_meta = AccountMeta(
        pubkey=ingl_constants.TEAM_ADDRESS, is_signer=False, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )
    vote_program_meta = AccountMeta(
        pubkey=ingl_constants.VOTE_PROGRAM_ID, is_signer=False, is_writable=False
    )

    account_metas = [
        payer_meta,
        storage_account_meta,
        registered_authorized_withdrawer_meta,
        vote_account_meta,
        sysvar_clock_meta,
        pda_authorized_withdrawer_meta,
        escrow_account_meta,
        team_account_meta,
        # programs listed below
        system_program_meta,
        vote_program_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.Buy(log_level=log_level), account_metas
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(transaction, client, payer)


async def process_withdraw_rewards(
    authorized_withdrawer: KeypairInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    pda_authorized_withdrawer_key, _pda_aw_bump = Pubkey.find_program_address(
        [ingl_constants.PDA_AUTHORIZED_WITHDRAWER_SEED], get_market_program_id()
    )
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )

    storage_account_data = Storage.parse(
        (await client.get_account_info(storage_account_key)).value.data
    )
    vote_account_key = Pubkey(storage_account_data.vote_account)

    # Forming AccountMetas
    authorized_withdrawer_meta = AccountMeta(
        pubkey=authorized_withdrawer.pubkey, is_signer=True, is_writable=True
    )
    vote_account_meta = AccountMeta(
        pubkey=vote_account_key, is_signer=False, is_writable=True
    )
    pda_authorized_withdrawer_meta = AccountMeta(
        pubkey=pda_authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    vote_program_meta = AccountMeta(
        pubkey=ingl_constants.VOTE_PROGRAM_ID, is_signer=False, is_writable=False
    )

    account_metas = [
        authorized_withdrawer_meta,
        vote_account_meta,
        pda_authorized_withdrawer_meta,
        storage_account_meta,
        # programs listed below
        vote_program_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.WithdrawRewards(log_level=log_level),
            account_metas,
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(
        transaction, client, authorized_withdrawer
    )


async def process_request_mediation(
    payer: KeypairInput,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )

    # Forming AccountMetas
    payer_meta = AccountMeta(pubkey=payer.pubkey, is_signer=True, is_writable=True)
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )

    account_metas = [
        payer_meta,
        storage_account_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.RequestMediation(log_level=log_level),
            account_metas,
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(transaction, client, payer)


async def process_mediate(
    payer: KeypairInput,
    mediation_shares: dict,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )

    escrow_account_key, _escrow_bump = Pubkey.find_program_address(
        [ingl_constants.ESCROW_SEED], get_market_program_id()
    )

    storage_account_data = Storage.parse(
        (await client.get_account_info(storage_account_key)).value.data
    )

    authorized_withdrawer_key = Pubkey(storage_account_data.authorized_withdrawer)
    buyer_key = Pubkey(storage_account_data.purchase.buyer)

    # Forming AccountMetas
    payer_meta = AccountMeta(pubkey=payer.pubkey, is_signer=True, is_writable=True)
    authorized_withdrawer_meta = AccountMeta(
        pubkey=authorized_withdrawer_key, is_signer=False, is_writable=True
    )
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    buyer_account_meta = AccountMeta(
        pubkey=buyer_key, is_signer=False, is_writable=True
    )
    escrow_account_meta = AccountMeta(
        pubkey=escrow_account_key, is_signer=False, is_writable=True
    )
    team_account_meta = AccountMeta(
        pubkey=ingl_constants.TEAM_ADDRESS, is_signer=False, is_writable=True
    )

    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    account_metas = [
        payer_meta,
        authorized_withdrawer_meta,
        storage_account_meta,
        buyer_account_meta,
        escrow_account_meta,
        team_account_meta,
        # programs listed below
        system_program_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.ProcessMediation(
                log_level=log_level, mediation_shares=mediation_shares
            ),
            account_metas,
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(transaction, client, payer)


async def process_validate_secondary_item_transfers(
    authorized_withdrawer: KeypairInput,
    item_index: int,
    client: AsyncClient,
    log_level: int = 0,
) -> str:
    storage_account_key, _storage_bump = Pubkey.find_program_address(
        [ingl_constants.PROGRAM_STORAGE_SEED], get_market_program_id()
    )
    escrow_account_key, _escrow_bump = Pubkey.find_program_address(
        [ingl_constants.ESCROW_SEED], get_market_program_id()
    )

    storage_account_data = Storage.parse(
        (await client.get_account_info(storage_account_key)).value.data
    )
    buyer_key = Pubkey(storage_account_data.purchase.buyer)
    assert authorized_withdrawer.pubkey == Pubkey(
        storage_account_data.authorized_withdrawer
    ), "Authorized withdrawer signing is not the same as the one in storage account"

    # Forming AccountMetas
    buyer_meta = AccountMeta(pubkey=buyer_key, is_signer=False, is_writable=True)
    storage_account_meta = AccountMeta(
        pubkey=storage_account_key, is_signer=False, is_writable=True
    )
    escrow_account_meta = AccountMeta(
        pubkey=escrow_account_key, is_signer=False, is_writable=True
    )
    authorized_withdrawer_meta = AccountMeta(
        pubkey=authorized_withdrawer.pubkey, is_signer=True, is_writable=True
    )
    system_program_meta = AccountMeta(
        pubkey=system_program.ID, is_signer=False, is_writable=False
    )

    account_metas = [
        buyer_meta,
        storage_account_meta,
        escrow_account_meta,
        authorized_withdrawer_meta,
        # programs listed below
        system_program_meta,
    ]

    # Forming Transaction
    transaction = Transaction()
    transaction.add(
        build_market_instruction(
            MarketInstructionEnum.enum.ValidateSecondaryItemsTransfers(
                item_index=item_index, log_level=log_level
            ),
            account_metas,
        )
    )

    # Sign Transaction, Send Transaction, and Confirm Transaction, Return Transaction Signature
    return await sign_send_confirm_return_tx_as_link(
        transaction, client, authorized_withdrawer
    )
